==> rmonitor-2.0.3.tar.bz2: rmonitor-2.0.3/exec_shell.c <==
/*
** $Id: exec_shell.c,v 1.5 2020/09/11 21:16:41 lpitcher Exp lpitcher $
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <sys/select.h>
#include <syslog.h>
#include <limits.h>

#include "logging.h"
#include "globals.h"
#include "exec_shell.h"

/***********************************************************************************/

/*
** Private functions
*/
static void adjustChildFds(int *, int *);
static void logAll(int, int, char *, char *);
static size_t logLine(char *, size_t, char *, char *, int);

/*
** MACROs used in this module:
**	BUFFERSIZE	size of each buffer allocated for target config
**	EXITPREFIX	base part of path to rmonitor exit directory
*/
#define BUFFERSIZE	(PIPE_BUF)

#if !defined EXITPREFIX
#define EXITPREFIX	"/usr/local/libexec"
#endif

/***********************************************************************************/

/*
** doExec()	fork off child to run script from shell
**		capture child's stdout & stderr and log them to syslog
**
** Accepts:	a pointer to a char string holding the interpreter path
**		a pointer to a char string holding the target script path
**
** Updates:	syslog with startup, termination, stdout & stderr messages
**
** Returns:	RUNERROR if child cannot be run, or
**		child status code, if child was run
*/
int doExec(char *shell, char *script)
{
  int rc = EXIT_SUCCESS;
  pid_t child;
  int	child_status,
  	stdout_pipe[2],
	stderr_pipe[2];

  pipe(stdout_pipe);
  pipe(stderr_pipe);

  switch(child = fork())
  {
    case 0: /* child */
      adjustChildFds(stdout_pipe, stderr_pipe);
      puts("PROCESSING Begins"); fflush(stdout);

      execlp(shell,shell,script,NULL);
      fprintf(stderr,"PROCESSING Failed to launch %s\n",script); fflush(stderr); 
      abort();	/* child will SIGABRT, parent will log as RUNERROR */
      
    case -1: /* fork failed */
      syslog(SYSLOG_LEVEL_ERROR,"JOB processing resource failure");
      rc = RUNERROR;
      break;

    default: /* parent */
      close(stdout_pipe[1]); close(stderr_pipe[1]);
      openlog(JobScript,LOG_CONS|LOG_PID,SYSLOG_FACILITY);
      logAll(stdout_pipe[0],stderr_pipe[0],NULL,NULL);
      openlog(JobName,LOG_CONS|LOG_PID,SYSLOG_FACILITY);

      waitpid(child,&child_status,0);
      rc = WIFEXITED(child_status) ? WEXITSTATUS(child_status) : RUNERROR;

      syslog(SYSLOG_LEVEL_MESSAGE,"PROCESSING Status %d",rc);
      break;
  }

  return rc;
}

/*
** doExit() selects (based on the lockname) a script in /usr/local/libexec/rmonitor
** and runs it, passing it the order parameter.
**
** Accepts:	a pointer to a char string identifying the resource exit
**		a pointer to a char string identifying the exit function
**
** Updates:	syslog with startup, termination, stdout & stderr messages
**
** Returns:	RUNERROR if exit script failed to execute
**		script exit status (0 to 127) if exit script was executed
*/
int doExit(char *lockname, char *order)
{
  char *exitpath;
  int rc;

  if (lockname && strlen(lockname) && order && strlen(order))
  {
    if (exitpath = malloc(FILENAME_MAX + strlen(lockname) + 1))
    {
      if (sprintf(exitpath,"%s/rmonitor/%s.exit",EXITPREFIX,lockname) <= FILENAME_MAX)
      {
        pid_t	child;
        int	child_status,
        	stdout_pipe[2],
		stderr_pipe[2];

        pipe(stdout_pipe);
        pipe(stderr_pipe);

        switch(child = fork())
        {
          case 0: /* child */
            adjustChildFds(stdout_pipe, stderr_pipe);
            puts("EXIT Begins"); fflush(stdout);

            execl(exitpath,exitpath,order,NULL);
            fprintf(stderr,"EXIT failed to launch %s\n",exitpath); fflush(stderr);
            abort();	/* child will SIGABRT, parent will log as RUNERROR */

          case -1: /* parent, fork failed OR child, execlp failed */
            syslog(SYSLOG_LEVEL_ERROR,"%s %s exit %s failed to launch",lockname,order,exitpath);
            rc = RUNERROR;	/* cant run exitpath as fork failed */
            break;

          default: /* parent, fork succeeded */
            close(stdout_pipe[1]); close(stderr_pipe[1]);
            logAll(stdout_pipe[0],stderr_pipe[0],lockname,order);

            waitpid(child,&child_status,0);
            rc = WIFEXITED(child_status) ? WEXITSTATUS(child_status) : RUNERROR;

            syslog(SYSLOG_LEVEL_MESSAGE,"%s %s: EXIT Status  %d",order,lockname,rc);
            break;
        }
      }
      else rc = RUNERROR;	/* cant run exit as exitname doesnt fit */
      free(exitpath);
    }
    else rc = RUNERROR;	/* cant run exit as cant allocate space for exitname */
  }
  else rc = RUNERROR;	/* cant run exit as lockname or order value are bad */

  return rc;
}

/*
** adjustChildFds	close all fds other than stdin, stdout and stderr
**			NB: will attach write end of stdout_pipe to stdout
**			    and write end of stderr_pipe to stderr before
**			    closing files
** Accepts:	a pointer to a pair of fd's associated with the stdout pipe
**		a pointer to a pair of fd's associated with the stderr pipe
**
** Updates:	redirects stdin to /dev/null
**		redirects stdout to write side fd of stdout pipe
**		redirects stderr to write side fd of stderr pipe
**		closes all other file handles
**
** Returns:		n/a
*/
static void adjustChildFds(int *stdout_pipe, int *stderr_pipe)
{
  long int fn;

  close(STDIN_FILENO);			/* close old stdin */
  open("/dev/null",O_RDONLY);		/* redirect STDIN to /dev/null */

  close(stdout_pipe[0]);		/* close read end FD */ 
  dup2(stdout_pipe[1],STDOUT_FILENO);	/* redirect STDOUT */
  close(stdout_pipe[1]);		/* close write end FD */

  close(stderr_pipe[0]);		/* close read end FD */ 
  dup2(stderr_pipe[1],STDERR_FILENO);	/* redirect STDERR */
  close(stderr_pipe[1]);		/* close write end FD */

  for (fn = sysconf(_SC_OPEN_MAX) - 1; fn > STDERR_FILENO; --fn)
    close(fn);
}

/*
** logAll()	syslog intercept for child stdout, stderr output
**		This function executes until all data from the
**		child process' stdout and stderr streams have
**		been intercepted, demarshalled into single lines,
**		and written to syslog.
**
** Accepts:	the input fd receiving the child processes stdout
**		the input fd receiving the child processes stderr
**		a string identifying the resource exit being logged, or NULL
**		a string identifying the function of the resource exit, or NULL
**
** Updates:	syslog with all received stdout & stderr data
**
** Returns:	n/a
*/
static void logAll(int outfd, int errfd, char *id, char *action)
{
  int	outData = 1,	/* 0 when read(outfd) returns 0 for EOF	*/
	errData = 1;	/* 0 when read(errfd) returns 0 for EOF	*/
  char	outBuf[BUFFERSIZE + 1],	/* line demarshalling buffer for outfd	*/
	errBuf[BUFFERSIZE + 1];	/* line demarshalling buffer for errfd	*/
  size_t outLen = 0,	/* number of bytes pending output from eutBuf */
	 errLen = 0;	/* number of bytes pending output from errBuf */

  fd_set	rfds;
  int		nfds, maxfd;
  ssize_t	nbytes;
  struct timeval waittime;

  while (outData || outLen || errData || errLen)
  {
    /* retrieve data from files */
    nfds = 0; maxfd = 0; FD_ZERO(&rfds);
    waittime.tv_sec = 0; waittime.tv_usec = 10;

    if (outData)
    {
      ++nfds;
      maxfd = (maxfd > outfd ? maxfd : outfd);
      FD_SET(outfd,&rfds);
    }
    if (errData)
    {
      ++nfds;
      maxfd = (maxfd > errfd ? maxfd : errfd);
      FD_SET(errfd,&rfds);
    }
    if (nfds > 0)
    {
      if (select(maxfd+1,&rfds,NULL,NULL,&waittime) > 0)
      {
        if (outData && FD_ISSET(outfd,&rfds))
        {
          nbytes = read(outfd,outBuf+outLen,sizeof(outBuf)-1-outLen);
          if (nbytes)
            outLen += nbytes;
          else
            outData = 0;
        }
        if (errData && FD_ISSET(errfd,&rfds))
        {
          nbytes = read(errfd,errBuf+errLen,sizeof(errBuf)-1-errLen);
          if (nbytes)
            errLen += nbytes;
          else
            errData = 0;
        }
      }
    }

    /* write lines from buffers */
    if (outLen) outLen = logLine(outBuf,outLen,id,action,SYSLOG_LEVEL_STDOUT);
    if (errLen) errLen = logLine(errBuf,errLen,id,action,SYSLOG_LEVEL_STDERR);
  }
}

/*
** logLine()	write one line of data to syslog. stop at first newline,
**		or at end of buffer, whichever comes first
**
** Accepts:	a pointer to a buffer holding the data to be logged
**		the length of the buffer
**		a string identifying the resource exit being logged, or NULL
**		a string identifying the function of the resource exit, or NULL
**		the syslog log level under which the data should be logged to syslog
**
** Updates:	syslog with all received stdout & stderr data
**		the passed buffer, deleting written data
**
** Returns:	number of bytes left in buffer after write
*/
static size_t logLine(char *buf, size_t len, char *id, char *action, int loglevel)
{
  static int logged = 1;
  size_t nl;

  /* write from [0] to first newline */
  for (nl = 0; nl < len && buf[nl] != '\n'; ++nl) {/*nop*/}
  buf[nl] = 0;

  if ((nl  > 0)					/* log text	*/
  ||  (nl == 0 && LogBlank == 1 && logged == 0)	/* log first blank line */
  ||  (nl == 0 && LogBlank == 2) )		/* log all blank lines	*/
  {
    logged = (nl > 0 ? 0 : 1);

    if (id && *id && action && *action)
      syslog(loglevel,"%s %s: %s",action,id,buf);
    else
      syslog(loglevel,"%s",buf);
  }

  /* pack buffer */
  if (len > nl + 1)
  {
    len = len - nl - 1;
    memmove(buf,buf + nl + 1, len);
  }
  else len = 0;

  return len;
}

/*
** Copyright Lew Pitcher (lew.pitcher@digitalfreehold.ca) February, 2013
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program (see the licence.txt file); if not, write to the Free
** Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
** 02110-1301 USA.
*/

==> rmonitor-2.0.4.tar.bz2: rmonitor-2.0.4/exec_shell.c <==
/*
** $Id: exec_shell.c,v 1.8 2020/11/17 17:50:49 lpitcher Exp $
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <syslog.h>
#include <limits.h>
#include <errno.h>

#include "logging.h"
#include "globals.h"
#include "exec_shell.h"

/***********************************************************************************/

/*
** Private functions
*/
static void adjustChildFds(int *, int *);
static void logAll(int, int, char *, char *);
static size_t logLine(char *, size_t, char *, char *, int);

/*
** MACROs used in this module:
**	BUFFERSIZE	size of each buffer allocated for target config
**	EXITPATH	string naming path to rmonitor exit scripts
*/
#define BUFFERSIZE	(PIPE_BUF)
#define	EXITPATH	"/usr/local/libexec/rmonitor"

/***********************************************************************************/

/*
** doExec()	fork off child to run script from shell
**		capture child's stdout & stderr and log them to syslog
**
** Accepts:	a pointer to a char string holding the interpreter path
**		a pointer to a char string holding the target script path
**
** Updates:	syslog with startup, termination, stdout & stderr messages
**
** Returns:	RUNERROR if child cannot be run, or
**		child status code, if child was run
*/
int doExec(char *shell, char *script)
{
  int rc = EXIT_SUCCESS;
  pid_t child;
  int	child_status,
  	stdout_pipe[2],
	stderr_pipe[2];

  pipe(stdout_pipe);
  pipe(stderr_pipe);

  switch(child = fork())
  {
    case 0: /* child */
      adjustChildFds(stdout_pipe, stderr_pipe);
      puts("PROCESSING Begins"); fflush(stdout);

      execlp(shell,shell,script,NULL);
      fprintf(stderr,"PROCESSING Failed to launch %s (%s)\n",script,strerror(errno));
      fflush(stderr); 
      abort();	/* child will SIGABRT, parent will log as RUNERROR */
      
    case -1: /* fork failed */
      syslog(SYSLOG_LEVEL_ERROR,"JOB resource failure: %s",strerror(errno));
      rc = RUNERROR;
      break;

    default: /* parent */
      close(stdout_pipe[1]); close(stderr_pipe[1]);
      openlog(JobScript,LOG_CONS|LOG_PID,SYSLOG_FACILITY);
      logAll(stdout_pipe[0],stderr_pipe[0],NULL,NULL);

      openlog(JobName,LOG_CONS|LOG_PID,SYSLOG_FACILITY);
      waitpid(child,&child_status,0);
      if (WIFEXITED(child_status))
      {
	rc = WEXITSTATUS(child_status);
	syslog(SYSLOG_LEVEL_MESSAGE,"PROCESSING Status %d",rc);
      }
      else
      {
	rc = RUNERROR;
	if (WIFSIGNALED(child_status))
	  syslog(SYSLOG_LEVEL_MESSAGE,"PROCESSING Killed: %s",strsignal(WTERMSIG(child_status)));
	else
	  syslog(SYSLOG_LEVEL_MESSAGE,"PROCESSING Terminated");
      }
      break;
  }

  return rc;
}

/*
** doExit() selects (based on the lockname) a script in EXITPATH
** and runs it, passing it the order parameter.
**
** Accepts:	a pointer to a char string identifying the resource exit
**		a pointer to a char string identifying the exit function
**
** Updates:	syslog with startup, termination, stdout & stderr messages
**
** Returns:	RUNERROR if exit script failed to execute
**		script exit status (0 to 127) if exit script was executed
*/
int doExit(char *lockname, char *order)
{
  char *exitpath;
  int rc;

  if (lockname && strlen(lockname) && order && strlen(order))
  {
    if (exitpath = malloc(FILENAME_MAX + strlen(lockname) + 1))
    {
      if (sprintf(exitpath,"%s/%s.exit",EXITPATH,lockname) <= FILENAME_MAX)
      {
        pid_t	child;
        int	child_status,
        	stdout_pipe[2],
		stderr_pipe[2];

        pipe(stdout_pipe);
        pipe(stderr_pipe);

        switch(child = fork())
        {
          case 0: /* child */
            adjustChildFds(stdout_pipe, stderr_pipe);
            puts("EXIT Begins"); fflush(stdout);

            execl(exitpath,exitpath,order,NULL);
	    fprintf(stderr,"EXIT failed to launch %s (%s)\n",exitpath,strerror(errno));
	    fflush(stderr);
            abort();	/* child will SIGABRT, parent will log as RUNERROR */

          case -1: /* parent, fork failed OR child, execlp failed */
            syslog(SYSLOG_LEVEL_ERROR,
		   "%s %s exit %s failed to launch: %s",
		   lockname,order,exitpath,strerror(errno));
            rc = RUNERROR;	/* cant run exitpath as fork failed */
            break;

          default: /* parent, fork succeeded */
            close(stdout_pipe[1]); close(stderr_pipe[1]);
            logAll(stdout_pipe[0],stderr_pipe[0],lockname,order);

            waitpid(child,&child_status,0);
	    if (WIFEXITED(child_status))
	    {
	      rc = WEXITSTATUS(child_status);
              syslog(SYSLOG_LEVEL_MESSAGE,"%s %s: EXIT Status %d",order,lockname,rc);
	    }
            else
            {
	      rc = RUNERROR;
	      if (WIFSIGNALED(child_status))
		syslog(SYSLOG_LEVEL_MESSAGE,
		       "%s %s: EXIT Killed: %s",
		       order,lockname,strsignal(WTERMSIG(child_status)));
	      else
	        syslog(SYSLOG_LEVEL_MESSAGE,
		       "%s %s: EXIT Terminated",
		        order,lockname);
	    }
            break;
        }
      }
      else rc = RUNERROR;	/* cant allocate space for exit path */
      free(exitpath);
    }
    else rc = RUNERROR;		/* cant allocate space for exitname */
  }
  else rc = RUNERROR;		/* bad lockname or order value */

  return rc;
}

/*
** adjustChildFds	close all fds other than stdin, stdout and stderr
**			NB: will attach write end of stdout_pipe to stdout
**			    and write end of stderr_pipe to stderr before
**			    closing files
** Accepts:	a pointer to a pair of fd's associated with the stdout pipe
**		a pointer to a pair of fd's associated with the stderr pipe
**
** Updates:	redirects stdin to /dev/null
**		redirects stdout to write side fd of stdout pipe
**		redirects stderr to write side fd of stderr pipe
**		closes all other file handles
**
** Returns:		n/a
*/
static void adjustChildFds(int *stdout_pipe, int *stderr_pipe)
{
  struct rlimit maxfds;

  /* redirect STDIN, STDOUT and STDERR */
  dup2(open("/dev/null",O_RDONLY),STDIN_FILENO);
  dup2(stdout_pipe[1],STDOUT_FILENO);
  dup2(stderr_pipe[1],STDERR_FILENO);

  /* close ALL other file descriptors */
  getrlimit(RLIMIT_NOFILE,&maxfds);
  while (--maxfds.rlim_max > STDERR_FILENO)
    close(maxfds.rlim_max);
}

/*
** logAll()	syslog intercept for child stdout, stderr output
**		This function executes until all data from the
**		child process' stdout and stderr streams have
**		been intercepted, demarshalled into single lines,
**		and written to syslog.
**
** Accepts:	the input fd receiving the child processes stdout
**		the input fd receiving the child processes stderr
**		a string identifying the resource exit being logged, or NULL
**		a string identifying the function of the resource exit, or NULL
**
** Updates:	syslog with all received stdout & stderr data
**
** Returns:	n/a
*/
static void logAll(int outfd, int errfd, char *id, char *action)
{
  int	outData = 1,	/* 0 when read(outfd) returns 0 for EOF	*/
	errData = 1;	/* 0 when read(errfd) returns 0 for EOF	*/
  char	outBuf[BUFFERSIZE + 1],	/* line demarshalling buffer for outfd	*/
	errBuf[BUFFERSIZE + 1];	/* line demarshalling buffer for errfd	*/
  size_t outLen = 0,	/* number of bytes pending output from eutBuf */
	 errLen = 0;	/* number of bytes pending output from errBuf */

  fd_set	rfds;
  int		nfds, maxfd;
  ssize_t	nbytes;
  struct timeval waittime;

  while (outData || outLen || errData || errLen)
  {
    /* retrieve data from files */
    nfds = 0; maxfd = 0; FD_ZERO(&rfds);
    waittime.tv_sec = 0; waittime.tv_usec = 10;

    if (outData)
    {
      ++nfds;
      maxfd = (maxfd > outfd ? maxfd : outfd);
      FD_SET(outfd,&rfds);
    }
    if (errData)
    {
      ++nfds;
      maxfd = (maxfd > errfd ? maxfd : errfd);
      FD_SET(errfd,&rfds);
    }
    if (nfds > 0)
    {
      if (select(maxfd+1,&rfds,NULL,NULL,&waittime) > 0)
      {
        if (outData && FD_ISSET(outfd,&rfds))
        {
          nbytes = read(outfd,outBuf+outLen,sizeof(outBuf)-1-outLen);
          if (nbytes)
            outLen += nbytes;
          else
            outData = 0;
        }
        if (errData && FD_ISSET(errfd,&rfds))
        {
          nbytes = read(errfd,errBuf+errLen,sizeof(errBuf)-1-errLen);
          if (nbytes)
            errLen += nbytes;
          else
            errData = 0;
        }
      }
    }

    /* write lines from buffers */
    if (outLen) outLen = logLine(outBuf,outLen,id,action,SYSLOG_LEVEL_STDOUT);
    if (errLen) errLen = logLine(errBuf,errLen,id,action,SYSLOG_LEVEL_STDERR);
  }
}

/*
** logLine()	write one line of data to syslog. stop at first newline,
**		or at end of buffer, whichever comes first
**
** Accepts:	a pointer to a buffer holding the data to be logged
**		the length of the buffer
**		a string identifying the resource exit being logged, or NULL
**		a string identifying the function of the resource exit, or NULL
**		the syslog log level under which the data should be logged to syslog
**
** Updates:	syslog with all received stdout & stderr data
**		the passed buffer, deleting written data
**
** Returns:	number of bytes left in buffer after write
*/
static size_t logLine(char *buf, size_t len, char *id, char *action, int loglevel)
{
  static int logged = 1;
  size_t nl;

  /* write from [0] to first newline */
  for (nl = 0; nl < len && buf[nl] != '\n'; ++nl) {/*nop*/}
  buf[nl] = 0;

  if ((nl  > 0)					/* log text	*/
  ||  (nl == 0 && LogBlank == 1 && logged == 0)	/* log first blank line */
  ||  (nl == 0 && LogBlank == 2) )		/* log all blank lines	*/
  {
    logged = (nl > 0 ? 0 : 1);

    if (id && *id && action && *action)
      syslog(loglevel,"%s %s: %s",action,id,buf);
    else
      syslog(loglevel,"%s",buf);
  }

  /* pack buffer */
  if (len > nl + 1)
  {
    len = len - nl - 1;
    memmove(buf,buf + nl + 1, len);
  }
  else len = 0;

  return len;
}

/*
** Copyright Lew Pitcher (lew.pitcher@digitalfreehold.ca) 2013, 2015, 2018, 2020
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program (see the licence.txt file); if not, write to the Free
** Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
** 02110-1301 USA.
*/
